#!/usr/bin/env python

### BEGIN INIT INFO
# Provides:          tinker-access-client
# Required-Start:    $remote_fs $syslog $network
# Required-Stop:     $remote_fs $syslog
# Default-Start:     2 3 4 5
# Default-Stop:      0 1 6
# Short-Description: tinker-access-client
# Description:       The tinker-access-client service is responsible for coordinating communication
# between RPi's modules (i.e. RFID reader, LCD, power relays etc..) and the remote tinker_access_server.
### END INIT INFO

import sys
import logging
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientDaemon import ClientDaemon
from CommandHandler import CommandHandler
from ClientOptionParser import ClientOptionParser, Command


def run():
    logger = ClientLogger.setup()
    try:
        with CommandHandler() as handler:
            handler.on(Command.STOP, ClientDaemon.stop)
            handler.on(Command.START, ClientDaemon.start)
            handler.on(Command.STATUS, ClientDaemon.status)
            handler.on(Command.RESTART, ClientDaemon.restart)

            #handler.on(Command.UPDATE, ClientDaemon.update)
            # TODO: UNINSTALL, RE-CONFIGURE, GET-CONFI-VALUE, SET-CONFIG-VALUE, GET-CONFIGS

            (opts, args) = ClientOptionParser().parse_args()
            handler.wait(opts, args)

    except (KeyboardInterrupt, SystemExit):
        pass

    except Exception as e:
        logger.debug('%s failed to handle the command.', PackageInfo.pip_package_name)
        logger.exception(e)
        sys.stdout.write(str(e))
        sys.stdout.flush()
        sys.exit(1)

    finally:
        logging.shutdown()

if __name__ == '__main__':
    run()



import json
import errno
import socket
from socket import error as socket_error

from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientOptionParser import ClientOptionParser, ClientOption


class ClientSocket(object):
    def __init__(self, timeout=1.5):
        self.__logger = ClientLogger.setup()
        self.__opts = ClientOptionParser().parse_args()[0]
        self.__timeout = timeout

    def __enter__(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(self.__timeout)
        self.socket.connect(('', self.__opts.get(ClientOption.CLIENT_PORT)))
        return self

    def send(self, opts, args):
        response = None
        try:
            self.socket.send(json.dumps({'opts': opts, 'args': args}))
            response = self.socket.recv(1024)
            self.socket.shutdown(socket.SHUT_RDWR)

        except socket_error as e:
            if e.errno != errno.ENOTCONN:
                raise e

        return response

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket.close()
        if exc_type:
            return False





# https://sourceforge.net/p/raspberry-gpio-python/code/ci/default/tree/source/c_gpio.h

# define MODE_UNKNOWN -1
# define BOARD        10
# define BCM          11
# define SERIAL       40
# define SPI          41
# define I2C          42
# define PWM          43

# define SETUP_OK           0
# define SETUP_DEVMEM_FAIL  1
# define SETUP_MALLOC_FAIL  2
# define SETUP_MMAP_FAIL    3
# define SETUP_CPUINFO_FAIL 4
# define SETUP_NOT_RPI_FAIL 5

# define INPUT  1 // is really 0 for control register!
# define OUTPUT 0 // is really 1 for control register!
# define ALT0   4

# define PUD_OFF  0
HIGH = 1
LOW = 0
BCM = 11
IN = 0
OUT = 1
PUD_DOWN = 1
PUD_UP = 2

# TODO: confirm these values
RISING = 1
FALLING = 2
BOTH = 3

FOO = 'BAR'

def setmode(self, *args):
    pass


def cleanup():
    pass


# noinspection PyPep8Naming
def setwarnings(self, *args):
    pass


def setup(self, *args):
    pass


def add_event_detect(self, *args, **kwargs):
    pass


def add_event_callback(self, *args, **kwargs):
    pass


def output(self, *args, **kwargs):
    pass


import os

tmp_file = '/tmp/login.txt'


class _VirtualSerial(object):
    def __init__(self, *args):
        self.__badge_code = None

    # noinspection PyPep8Naming
    def flushInput(self, *args):
        if os.path.isfile(tmp_file):
            os.remove(tmp_file)
            self.__badge_code = None

    # noinspection PyPep8Naming
    def flushOutput(self, *args):
        pass

    def inWaiting(self, *args):
        self.__read_line()
        return len(self.__badge_code) if self.__badge_code is not None else 0

    def readline(self, *args):
        print 'badge_code: ' + self.__badge_code
        return self.__badge_code

    # noinspection PyMethodMayBeStatic
    def __read_line(self):
        if os.path.isfile(tmp_file):
            with open(tmp_file, 'r') as f:
                badge_code = f.readline()
                if len(badge_code):
                    self.__badge_code = badge_code


# noinspection PyClassHasNoInit
class VirtualSerial:
    Serial = _VirtualSerial



import os
import sys
import time
from subprocess import \
    call, \
    check_output, \
    CalledProcessError

from Client import Client
from Command import Command
from ClientSocket import ClientSocket
from daemonize import Daemonize
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientOption import ClientOption


# noinspection PyClassHasNoInit
class ClientDaemon:

    @staticmethod
    def start(opts, args):
        logger = ClientLogger.setup()
        if not ClientDaemon.__status(opts, args):
            try:
                client = Client()
                pid_file = opts.get(ClientOption.PID_FILE)
                foreground = opts.get(ClientOption.DEBUG)
                daemon = Daemonize(
                    app=PackageInfo.pip_package_name,
                    pid=pid_file,
                    action=client.run,
                    foreground=foreground,
                    verbose=True,
                    logger=logger,
                    auto_close_fds=False
                )
                daemon.start()
            except Exception as e:
                logger.debug('%s start failed.', PackageInfo.pip_package_name)
                logger.exception(e)
                raise e
        else:
            sys.stdout.write('{0} is already running...\n'.format(PackageInfo.pip_package_name))
            sys.stdout.flush()
            sys.exit(1)

    @staticmethod
    def stop(opts, args):
        logger = ClientLogger.setup()
        pid_file = opts.get(ClientOption.PID_FILE)
        logout_coast_time = opts.get(ClientOption.LOGOUT_COAST_TIME)
        try:
            # Wait for what we assume will be a graceful exit
            ClientDaemon.__stop(opts, args)
            remaining_logout_coast_time = logout_coast_time
            while remaining_logout_coast_time and os.path.isfile(pid_file):
                remaining_logout_coast_time -= 1
                time.sleep(1)

            # if any processes still exists at this point... we will become more persuasive...
            for pid in check_output(['pgrep', '-f', '{0} start'.format(
                    PackageInfo.pip_package_name)]).splitlines():
                try:
                    call(['kill', '-9', pid])
                except Exception as e:
                    logger.exception(e)
        except CalledProcessError:
            pass

        # if the pid file still exist at this point... nuke it from orbit!
        if os.path.isfile(pid_file):
            try:
                os.remove(pid_file)
            except Exception as e:
                logger.exception(e)

    @staticmethod
    def restart(opts, args):
        logger = ClientLogger.setup()
        try:
            args[0] = Command.STOP.get('command')
            ClientDaemon.stop(opts, args)
            args[0] = Command.START.get('command')
            ClientDaemon.start(opts, args)
        except Exception as e:
            logger.debug('%s restart failed.', PackageInfo.pip_package_name)
            logger.exception(e)
            raise e

    # @staticmethod
    # def update(opts, args):
    #     pass

    # noinspection PyUnusedLocal
    @staticmethod
    def status(opts, args):
        logger = ClientLogger.setup()
        status = ClientDaemon.__status(opts, args)
        if status:
            sys.stdout.write('Status: {0}\n'.format(status))
            sys.stdout.flush()
            sys.exit(0)

        sys.stdout.write('the {0} is unavailable...\n'
                         'It may need to be restarted, try: \'sudo {0} restart\'\n'
                         .format(PackageInfo.pip_package_name))
        sys.stdout.flush()
        sys.exit(1)

    @staticmethod
    def __status(opts, args):
        # noinspection PyBroadException
        try:
            with ClientSocket() as socket:
                return socket.send(opts, args)
        except Exception:
            pass

    @staticmethod
    def __stop(opts, args):
        # noinspection PyBroadException
        try:
            with ClientSocket() as socket:
                socket.send(opts, args)
        except Exception:
            pass




from ClientLogger import ClientLogger
from LoggedRequest import LoggedRequest
from UserRegistrationException import UserRegistrationException
from UnauthorizedAccessException import UnauthorizedAccessException

#TODO: rename this to a more meaningful name, i.e. RemoteServer, TinkerAccessServerApi etc..
class ServerApi(object):
    def __init__(self, server_address, device_id):
        self.__logger = ClientLogger.setup()
        self.__server_address = server_address
        self.__device_id = device_id

    def login(self, user_badge_code):
        try:
            url = "%s/device/%s/code/%s" % (self.__server_address, self.__device_id, user_badge_code)
            response = LoggedRequest.get(url)

        except Exception as e:
            self.__logger.debug('Login attempt failed due to unexpected exception.')
            raise e

        else:
            data = response.json()
            login_response = {
                'user_name': data.get('username'),
                'device_name': data.get('devicename'),
                'user_id': data.get('userid'),
                'badge_code': user_badge_code,
                'remaining_seconds': data.get('time', 0) * 60
            }

            if login_response.get('remaining_seconds') > 0:
                self.__logger.debug('Login succeeded.')
                return login_response

        self.__logger.debug('Login attempt failed. The user is not authorized for this device.')
        raise UnauthorizedAccessException

    def logout(self, user_id):
        try:
            url = "%s/device/%s/logout/%s" % (self.__server_address, self.__device_id, user_id)
            LoggedRequest.get(url)

        except Exception as e:
            self.__logger.debug('Logout attempt failed due to unexpected exception.')
            raise e

        self.__logger.debug('Logout succeeded.')

    def register_user(self, trainer_id, trainer_badge_code, user_badge_code):
        try:
            url = "%s/admin/marioStar/%s/%s/%s/%s" % (
                self.__server_address, trainer_id, trainer_badge_code, self.__device_id, user_badge_code)
            response = LoggedRequest.get(url)

        except Exception as e:
            self.__logger.debug(
                'User registration failed due to unexpected exception.')
            raise e

        else:
            if response.text() == 'true':
                self.__logger.debug('User registration succeeded.')
                return

        self.__logger.debug('User registration failed. The user is not authorized for this device.')
        raise UserRegistrationException



import time
import json
import threading
from ClientLogger import ClientLogger
from ClientOptionParser import ClientOption


class Channel(object):
    LCD, SERIAL, LED, PIN = range(0, 4)

    def __new__(cls, channel):
        for key, value in vars(Channel).items():
            if not key.startswith('__'):
                if value == channel:
                    return key
        return None


class DeviceApi(object):
    def __init__(self, opts):
        self.__opts = opts
        self.__fault = False
        self.__should_exit = False
        self.__edge_detected = False
        self.__logger = ClientLogger.setup()

        try:
            # !Important: These modules are imported locally at runtime so that the unit test
            # can run on non-rpi devices where GPIO won't load and cannot work
            # (i.e. the build server, and dev environments) see: test_DeviceApi.py
            self.__init__device_modules()

        except (RuntimeError, ImportError) as e:
            self.__logger.exception(e)
            raise e

        except Exception as e:
            self.__logger.debug('Device initialization failed with %s.', json.dumps(opts, indent=4, sort_keys=True))
            self.__logger.exception(e)
            raise e

    def __init__device_modules(self):
        self.__init_gpio()
        self.__init_lcd()
        self.__init__serial()

    def __init_gpio(self):
        try:
            import RPi.GPIO as GPIO

            # TODO: log print board info after init...
            # To discover information about your RPi:
            # GPIO.RPI_INFO
            # To discover the Raspberry Pi board revision:
            # GPIO.RPI_INFO['P1_REVISION']
            # GPIO.RPI_REVISION    (deprecated)
            # To discover the version of RPi.GPIO:
            # GPIO.VERSION

        except Exception as e:
            try:
                import debug.VirtualRPi.VirtualGPIO as GPIO
            except Exception as ex:
                self.__logger.debug('Failed to patch RPi.GPIO module with virtual GPIO module.')
                self.__logger.exception(ex)
                raise e

        GPIO.setmode(GPIO.BCM)

        # TODO: remove once lcdModule is fixed fixed to not also call cleanup
        # GPIO.cleanup()
        # GPIO.setwarnings(False)

        GPIO.setup(self.__opts.get(ClientOption.PIN_LED_RED), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LED_BLUE), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LED_GREEN), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_POWER_RELAY), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LOGOUT), GPIO.IN, GPIO.PUD_DOWN)
        GPIO.setup(self.__opts.get(ClientOption.PIN_CURRENT_SENSE), GPIO.IN, GPIO.PUD_DOWN)
        self.GPIO = GPIO

    def __init_lcd(self):
        try:
            import lcdModule.LCD as LCD
        except Exception as e:
            try:
                import debug.VirtualLcd.VirtualLCD as LCD
            except Exception as ex:
                self.__logger.debug('Failed to patch lcdModule.LCD module with virtual LCD module.')
                self.__logger.exception(ex)
                raise e

        LCD.lcd_init()
        self.__LCD = LCD

    def __init__serial(self):
        serial_port_name = self.__opts.get(ClientOption.SERIAL_PORT_NAME)
        serial_port_speed = self.__opts.get(ClientOption.SERIAL_PORT_SPEED)

        try:
            import serial
            self.__serial_connection = serial.Serial(serial_port_name, serial_port_speed)

        except Exception as e:
            try:
                # noinspection PyPep8Naming
                from debug.VirtualSerial import VirtualSerial as serial
                self.__serial_connection = serial.Serial(serial_port_name, serial_port_speed)
            except Exception as ex:
                self.__logger.debug('Failed to patch serial module with virtual LCD module.')
                self.__logger.exception(ex)
                raise e

        self.__serial_connection.flushInput()
        self.__serial_connection.flushOutput()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__do_cleanup()

    def __do_cleanup(self):
        self.GPIO.cleanup()

    def __do_callback(self, call_back, *args, **kwargs):
        try:
            call_back(*args, **kwargs)
        except Exception as e:
            self.__logger.debug('Failed to execute callback.')
            self.__logger.exception(e)
        finally:
            self.__edge_detected = True

    def __poll_for_serial_input(self, call_back):
        while not self.__should_exit:
            try:
                badge_code = self.read(Channel.SERIAL)
                if badge_code:
                    self.__do_callback(call_back, badge_code=badge_code)
            except Exception as e:
                self.__logger.exception(e)
            time.sleep(.5)

    def __read_from_serial(self):
        serial_connection = self.__serial_connection

        if serial_connection.inWaiting() > 1:
            value = serial_connection.readline().strip()[-12:]
            serial_connection.flushInput()
            serial_connection.flushOutput()
            return value

        return None

    def __read_from_pin(self, pin, expected_state):
        # noinspection PyPep8Naming
        GPIO = self.GPIO
        expected_state = GPIO.LOW if not expected_state else GPIO.HIGH
        return GPIO.input(pin) == expected_state

    def __write_to_led(self, red, green, blue):
        # noinspection PyPep8Naming
        GPIO = self.GPIO
        GPIO.output(self.__opts.get(ClientOption.PIN_LED_RED), red)
        GPIO.output(self.__opts.get(ClientOption.PIN_LED_GREEN), green)
        GPIO.output(self.__opts.get(ClientOption.PIN_LED_BLUE), blue)

    def __write_to_lcd(self, first_line, second_line):
        # noinspection PyPep8Naming
        LCD = self.__LCD
        LCD.lcd_string(first_line, LCD.LCD_LINE_1)
        LCD.lcd_string(second_line, LCD.LCD_LINE_2)

    def __write_to_pin(self, pin, state):
        # noinspection PyPep8Naming
        GPIO = self.GPIO
        state = GPIO.LOW if not state else GPIO.HIGH
        GPIO.output(pin, state)

    def on(self, channel, **kwargs):
        pin = kwargs.get('pin')
        direction = kwargs.get('direction')
        call_back = kwargs.get('call_back')

        # noinspection PyPep8Naming
        GPIO = self.GPIO
        if channel is Channel.PIN and pin and direction and call_back:

            def edge_detected(*args, **kwargs):
                self.__do_callback(call_back, args, kwargs)

            GPIO.add_event_detect(pin, direction, callback=edge_detected, bouncetime=250)

        elif channel is Channel.SERIAL and direction is self.GPIO.IN and call_back:
            poll_for_serial_input = threading.Thread(
                name='poll_for_serial_input',
                target=self.__poll_for_serial_input,
                args=(call_back,)
            )
            poll_for_serial_input.daemon = True
            poll_for_serial_input.start()

        else:
            raise NotImplementedError

    def read(self, channel, *args):
        if self.__fault:
            return

        channel_name = Channel(channel)
        try:

            if channel == Channel.SERIAL:
                value = self.__read_from_serial()

            elif channel == Channel.PIN:
                pin = args[0] if len(args) >= 1 else None
                expected_state = args[1] if len(args) >= 2 else True
                value = self.__read_from_pin(pin, expected_state)

            else:
                raise NotImplementedError

        except Exception as e:
            self.__logger.debug('Read from %s failed with args \'%s\'.', channel_name, args)
            self.__logger.exception(e)
            raise e

        return value

    def write(self, channel, *args):
        if self.__fault:
            return

        channel_name = Channel(channel)
        try:
            if channel == Channel.LED:
                red = len(args) >= 1 and args[0] is True
                green = len(args) >= 2 and args[1] is True
                blue = len(args) >= 3 and args[2] is True
                self.__write_to_led(red, green, blue)

            elif channel == Channel.LCD:
                first_line = args[0] if len(args) >= 1 else ''
                second_line = args[1] if len(args) >= 2 else ''
                self.__write_to_lcd(first_line, second_line)

            elif channel == Channel.PIN:
                pin = args[0] if len(args) >= 1 else None
                state = args[1] if len(args) >= 2 else None
                self.__write_to_pin(pin, state)

            else:
                raise NotImplementedError

        except Exception as e:
            self.__logger.debug('Write to \'%s\' failed with args \'%s\'.', channel_name, args)
            self.__logger.exception(e)
            raise e

    def wait(self):
        while not self.__should_exit and not self.__edge_detected:
            time.sleep(1)
        self.__edge_detected = False

    def fault(self):
        self.__fault = True

    def stop(self):
        self.__should_exit = True



import json
import errno
import socket
import threading
from socket import error as socket_error

from Command import Command
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from CommandHandler import CommandHandler

# TODO: while technically functional, I plan to revisit this because there are a few things
# that can be improved i.e. use 'select' to wait for connections, self.__listener.accept blocks,
# which prevents the thread from exiting gracefully in the event of a non STOP command shutdown i.e. CTRL-C:


class RemoteCommandHandler(CommandHandler):

    def __init__(self, execute_initial_command=False):
        self.__client = None
        self.__listener = None
        self.__logger = ClientLogger.setup()
        super(RemoteCommandHandler, self).__init__(execute_initial_command)

    def __enter__(self):
        try:
            self.__listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            self.__listener.bind(('', 8089)) # TODO: add and use additional config value for this listener
            self.__listener.settimeout(None)
            self.__listener.listen(5)

            # TODO: lookup async select patterns with sockets in python, or other libraries to use
            # to reduce this boiler plate code
        except Exception as e:
            self.__logger.debug('Unable to establish the %s listener.', PackageInfo.pip_package_name)
            self.__logger.exception(e)
            #thread.interrupt_main()  # TODO: investigate further, causing the unit test to fail and builds to fail

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__listener:
            try:
                self.__listener.shutdown(socket.SHUT_RDWR)
            except socket_error as e:
                if e.errno != errno.ENOTCONN:
                    self.__logger.exception(e)
                    raise e

            except Exception as e:
                self.__logger.exception(e)
                raise e
            finally:
                self.__listener.close()

        super(RemoteCommandHandler, self).__exit__(exc_type, exc_val, exc_tb)

    def __listen(self):
        cmd = None
        while cmd is not Command.STOP:
            client = None

            try:
                (client, addr) = self.__listener.accept()
                data = json.loads(client.recv(1024))
                opts = data.get('opts')
                args = data.get('args')
                cmd = Command(args[0]) if len(args) >= 1 else None
                response = self.wait(opts, args)
                client.send(response or '')
                client.shutdown(socket.SHUT_RDWR)

            except socket_error as e:
                # TODO: ignored for now, need to work out some issues to exit gracefully
                # will probably replace .accept with .select
                pass
                # if e.errno != errno.ENOTCONN:
                #     self.__logger.exception(e)
                #     raise e

            except Exception as e:
                self.__logger.exception(e)
                raise e

            finally:
                if client:
                    client.close()

    def listen(self):
        t = threading.Thread(name='remote-command-handler', target=self.__listen)
        t.daemon = True
        t.start()




import time
import threading

from transitions import Machine
# TODO: Enable LockedHierarchicalGraphMachine https://github.com/tyarkoni/transitions#-extensions
# from transitions.extensions import LockedHierarchicalGraphMachine as Machine

from Command import Command
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from DeviceApi import DeviceApi, Channel
from RemoteCommandHandler import RemoteCommandHandler
from ClientOptionParser import ClientOptionParser, ClientOption

logout_timer_interval_seconds = 1


class State(object):
    IDLE = 'IDLE'
    IN_USE = 'IN_USE'
    INITIALIZED = 'INITIALIZED'
    IN_FAULT = 'IN_FAULT'


class Trigger(object):
    IDLE = 'idle'
    LOGIN = 'login'
    LOGOUT = 'logout'


class Client(Machine):
    def __init__(self):
        self.__device = None
        self.__user_info = None
        self.__logout_timer = None
        self.__should_exit = False
        self.__logger = ClientLogger.setup()
        self.__opts = ClientOptionParser().parse_args()[0]

        states = []
        for key, _ in vars(State).items():
            if not key.startswith('__'):
                states.append(key)

        transitions = [
            {
                'trigger': Trigger.IDLE,
                'source': State.INITIALIZED,
                'dest': State.IDLE
            },
            {
                'trigger': Trigger.LOGIN,
                'source': State.IDLE,
                'dest': State.IN_USE,
                'conditions': 'authorized'
            },

            # TODO: handle case where use re-badges in, or another users badges in while the machine is in use
            # {
            #     'trigger': Trigger.LOGIN,
            #     'source': State.IN_USE,
            #     'dest': State.IN_USE,
            #     'conditions': 're-authorized' #extend time if same user, otherwise go through login process
            # },

            {
                'trigger': Trigger.LOGIN,
                'source': State.IDLE,
                'dest': State.IDLE,
                'unless': 'authorized',
                'before': 'not_authorized'
            },
            {
                'trigger': Trigger.LOGOUT,
                'source': [State.IN_USE, State.IDLE],
                'dest': State.IDLE
            }
        ]

        Machine.__init__(self, states=states, transitions=transitions, initial=State.INITIALIZED)

    def __handle_stop_command(self, **kwargs):
        self.__stop()

    def __stop(self):
        # what if device is in use, should stop return error, or forcefully do a stop?
        # TODO: wait for client status(self.state) to be done before we exit...
        # i.e. logout should be complete etc...

        self.__ensure_idle()
        self.__device.stop()
        self.__exit()

    def __exit(self):
        self.__should_exit = True

    def __handle_status_command(self, **kwargs):
        return self.state

    def __enable_power(self):
        self.__device.write(Channel.PIN, self.__opts.get(ClientOption.PIN_POWER_RELAY), True)

    def __disable_power(self):
        self.__device.write(Channel.PIN, self.__opts.get(ClientOption.PIN_POWER_RELAY), False)

    def __do_logout(self):
        self.__cancel_logout_timer()

        if self.__user_info:
            pass
            # ServerApi.doLogout

        self.__user_info = None

    def __ensure_idle(self):
        self.__do_logout()
        self.__disable_power()
        self.__device.write(Channel.LED, False, False, True)

    def __ensure_in_use(self):
        self.__enable_power()
        self.__device.write(Channel.LED, False, True, False)

    def __logout_timer_tick(self, remaining_seconds):
        remaining_seconds -= logout_timer_interval_seconds

        if remaining_seconds <= 0:
            self.logout()
            return

        if remaining_seconds < 300:
            self.__device.write(Channel.LED, True, False, False)

        self.__device.write(Channel.LCD, 'Access granted', 'Time Remaining: {0}s'.format(remaining_seconds))

        self.__logout_timer = threading.Timer(
            logout_timer_interval_seconds,
            self.__logout_timer_tick,
            [remaining_seconds, ]
        )
        self.__logout_timer.start()

    def __cancel_logout_timer(self):
        if self.__logout_timer:
            self.__logout_timer.cancel()

    def __start_logout_timer(self, remaining_seconds):
        self.__cancel_logout_timer()
        self.__logout_timer = threading.Timer(
            logout_timer_interval_seconds,
            self.__logout_timer_tick,
            [remaining_seconds, ]
        )
        self.__logout_timer.start()

    def authorized(self, *args, **kwargs):
        if not self.__user_info:
            badge_code = kwargs.get('badge_code')
            if badge_code:
                self.__logger.debug('badge code detected... %s', badge_code)
                # ServerApi.doLogin
                if badge_code == 'bar':
                    self.__user_info = badge_code

        return self.__user_info is not None

    def not_authorized(self, *args, **kwargs):
        self.__device.write(Channel.LED, True, False, False)
        self.__device.write(Channel.LCD, 'Access Denied', 'Take the class')
        time.sleep(2)

    # noinspection PyPep8Naming
    def on_enter_IN_FAULT(self):
        self.__device.fault()

    # noinspection PyPep8Naming
    def on_enter_IN_USE(self, *args, **kwargs):
        self.__ensure_in_use()
        self.__device.write(Channel.LCD, 'Access granted')

        # TODO: include user_name, time_remaining, in Access granted for % s granted with time % s"
        # TODO: start timer for logout trigger

        remaining_seconds = 60  # use value returned from service call, total remaining time should be in seconds
        self.__start_logout_timer(remaining_seconds)

    # noinspection PyPep8Naming
    def on_enter_IDLE(self, *args, **kwargs):
        self.__ensure_idle()
        self.__device.write(Channel.LCD, 'Scan Badge', 'To Login')

    def __run(self):
        opts = self.__opts

        with RemoteCommandHandler() as handler, DeviceApi(opts) as device:
            self.__device = device

            handler.on(Command.STOP, self.__handle_stop_command)
            handler.on(Command.STATUS, self.__handle_status_command)
            handler.listen()

            device.on(
                Channel.SERIAL,
                direction=device.GPIO.IN,
                call_back=self.login
            )

            device.on(
                Channel.PIN,
                pin=opts.get(ClientOption.PIN_LOGOUT),
                direction=device.GPIO.RISING,
                call_back=self.logout
            )

            self.idle()
            while not self.__should_exit:
                self.__logger.debug('%s is waiting...', PackageInfo.pip_package_name)
                device.wait()

    def run(self):
        while not self.__should_exit:
            try:
                self.__run()

            except (KeyboardInterrupt, SystemExit):
                self.__exit()

            except Exception as e:
                self.set_state(State.IN_FAULT)
                self.__logger.debug('%s failed.', PackageInfo.pip_package_name)
                self.__logger.exception(e)

            if not self.__should_exit:
                self.__logger.debug('Retrying in 5 seconds...')
                time.sleep(5)
                self.set_state(State.INITIALIZED)



# noinspection PyBroadException
# noinspection PyUnresolvedReferences,PyClassHasNoInit


class PackageInfo:
    python_package_name = 'tinker_access_client'
    pip_package_name = python_package_name.replace('_', '-')
    version = None

    try:
        from version import __version__ as version
    except:
        pass



import sys
import time
import json
from Command import Command
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
# from ClientDaemon import ClientDaemon
from ClientOptionParser import ClientOptionParser


# noinspection PyClassHasNoInit
class CommandHandler(object):

    def __init__(self, execute_initial_command=True):
        self.__logger = ClientLogger.setup()
        self.__handlers = []

    # TODO: add test and implement... __enter__, __exit__
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    # def __del__(self):
    #     pass

    def wait(self, opts, args):
        cmd = Command(args[0].lower() if len(args) >= 1 and len(args[0].lower()) >= 1 else None)

        result = None
        if cmd and len(self.__handlers):
            for (command, call_back) in self.__handlers:
                if command is cmd:
                    try:
                        result = call_back(opts=opts, args=args)
                    except Exception as e:
                        self.__logger.exception(e)

        return result

    def on_command(self, call_back):
        # TODO: implement... probably will need another thread..
        pass

    def on(self, command, call_back):
        self.__handlers.append((command, call_back,))